/** @file
 * SASL-based authentication mechanisms for identd.
 *
 * @author Booker Bense <bbense@stanford.edu>
 * @date 1996-08-28
 *
 * $Id: ident_auth.c,v 1.7 2002/10/29 05:39:23 eagle Exp $
 */

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h> 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/socket.h> 
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>

#include <krb.h>

#include "identd.h" 

/* Compatiblity with older krb4 srcs */ 
#ifndef KRB_INT32
#define KRB_INT32 KRB4_32
#endif

/* Exported definition of responder authentication mechanism */
static int krb_responder_start(); 
static int null_responder_start(); 
static int krb_responder_auth(); 
static int null_responder_auth(); 
static void krb_free_state(); 
static void null_free_state(); 
static int trivial_responder_start(); 
static int trivial_responder_auth(); 
extern char *sfgets() ;

/* GSS-API implementation. */
extern int gssapi_responder_start();
extern int gssapi_responder_auth();
extern void gssapi_free_state();

struct ident_responder_auth krb_ident_auth = {
    "KERBEROS_V4",
    krb_responder_start,
    krb_responder_auth,
    krb_free_state
};

struct ident_responder_auth gssapi_ident_auth = {
    "GSSAPI",
    gssapi_responder_start,
    gssapi_responder_auth,
    gssapi_free_state
};

struct ident_responder_auth otp_ident_auth = {
    "OTP_RFC1938",
    null_responder_start,
    null_responder_auth,
    null_free_state
};

struct ident_responder_auth trivial_ident_auth = {
    "TRIVIAL",
    trivial_responder_start,
    trivial_responder_auth,
    null_free_state
};


#define IDENT_METHODS_COUNT 4
/* TRIVIAL method must always be the last method */ 
#define IDENT_TRIVIAL_METHOD (IDENT_METHODS_COUNT - 1 ) 
struct ident_responder_auth *auth_methods[IDENT_METHODS_COUNT] = { &krb_ident_auth, 
						&gssapi_ident_auth,
						&otp_ident_auth,
						&trivial_ident_auth
}; 

int 
parse_auth(line,auth_struct)
char *line;  
struct ident_responder_auth **auth_struct ; 
{ 
  int rcode ;
  int i ; 
  char auth_method[128] ;
  *auth_struct = NULL ; 
  if ( line == NULL ) { 
    *auth_struct = auth_methods[IDENT_TRIVIAL_METHOD] ; 
    return IDENT_AUTH_OKAY ; 
  }
  /* Grok to the next , */ 

  rcode = sscanf(line ," %32[^ \t\n\r,]",auth_method ); 
  

  for(i = 0 ; i < IDENT_METHODS_COUNT ; i++ ) { 
    if ( strncmp(auth_method,auth_methods[i]->auth_method,
		 strlen(auth_methods[i]->auth_method)) == 0 ) {
      *auth_struct = auth_methods[i] ; 
      return(IDENT_AUTH_OKAY) ; 
    }
  }
  return(IDENT_INVALID_REQ_INFO); 
  
  
} 


/* All the code from here to EOF is C&P from imclient.c in the imap.1.4
libraries. 
*/ 

#define XX 127
/*
 * Tables for encoding/decoding base64
 */
static const char basis_64[] =
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static const char index_64[256] = {
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,62, XX,XX,XX,63,
    52,53,54,55, 56,57,58,59, 60,61,XX,XX, XX,XX,XX,XX,
    XX, 0, 1, 2,  3, 4, 5, 6,  7, 8, 9,10, 11,12,13,14,
    15,16,17,18, 19,20,21,22, 23,24,25,XX, XX,XX,XX,XX,
    XX,26,27,28, 29,30,31,32, 33,34,35,36, 37,38,39,40,
    41,42,43,44, 45,46,47,48, 49,50,51,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
    XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX, XX,XX,XX,XX,
};
#define CHAR64(c)  (index_64[(unsigned char)(c)])

/*
 * Decode in-place the base64 data in 'input'.  Returns the length
 * of the decoded data, or -1 if there was an error.
 */
int 
ident_decodebase64(input)
char *input;
{
    int len = 0;
    unsigned char *output = (unsigned char *)input;
    int c1, c2, c3, c4;

    while (*input) {
	c1 = *input++;
	if (CHAR64(c1) == XX) return -1;
	c2 = *input++;
	if (CHAR64(c2) == XX) return -1;
	c3 = *input++;
	if (c3 != '=' && CHAR64(c3) == XX) return -1; 
	c4 = *input++;
	if (c4 != '=' && CHAR64(c4) == XX) return -1;
	*output++ = (CHAR64(c1) << 2) | (CHAR64(c2) >> 4);
	++len;
	if (c3 == '=') break;
	*output++ = ((CHAR64(c2) << 4) & 0xf0) | (CHAR64(c3) >> 2);
	++len;
	if (c4 == '=') break;
	*output++ = ((CHAR64(c3) << 6) & 0xc0) | CHAR64(c4);
	++len;
    }

    return len;
}
	

/*
 * Write to the string 'output' the base-64 encoded data
 * 'input', of (unencoded) length 'len'.
 */

void
ident_writebase64(input, output, len)
char *input;
char *output;
int len;
{
    char buf[IDENT_BUFFER_SIZE];
    char *outptr ; 
    int buflen = 0;
    int c1, c2, c3;
    outptr = output ; 
    while (len) {
	if (buflen >= sizeof(buf)-4) {
	    strncpy(outptr,buf,buflen);
	    buflen = 0;
	    outptr += buflen ; 
	}
	
	c1 = (unsigned char)*input++;
	buf[buflen++] = basis_64[c1>>2];

	if (--len == 0) c2 = 0;
	else c2 = (unsigned char)*input++;
	buf[buflen++] = basis_64[((c1 & 0x3)<< 4) | ((c2 & 0xF0) >> 4)];

	if (len == 0) {
	    buf[buflen++] = '=';
	    buf[buflen++] = '=';
	    break;
	}

	if (--len == 0) c3 = 0;
	else c3 = (unsigned char)*input++;

	buf[buflen++] = basis_64[((c2 & 0xF) << 2) | ((c3 & 0xC0) >>6)];
	if (len == 0) {
	    buf[buflen++] = '=';
	    break;
	}

	--len;
	buf[buflen++] = basis_64[c3 & 0x3F];
    }

    if (buflen >= sizeof(buf)-2) {
	strncpy(outptr, buf, buflen);
	outptr += buflen ; 
	buflen = 0;
    }
    buf[buflen++] = '\0';
    strcpy(outptr, buf);
}

/* 
**
**     Kerberos auth routines
**
**
*/ 


/* acte_krb.c -- KERBEROS_V4 authentication routines for IMAP.
 *
 *	(C) Copyright 1994 by Carnegie Mellon University
 *
 *                      All Rights Reserved
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of Carnegie
 * Mellon University not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  Carnegie Mellon University makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 */

extern char *lcase();
extern char *krb_get_phost(), *krb_realmofhost();

static char *srvtab = "";	/* Srvtab filename */

/* Maximum number of bytes of overhead the protection mechanisms use */
#define PROTECTION_OVERHEAD 31

/* Private state used by this mechanism */
struct krb_state {
    /* common */
    char service[MAX_K_NAME_SZ+1];
    int authstepno;
    des_cblock session;	/* Our session key */
    des_key_schedule schedule; /* Schedule for our session key */
    KRB_INT32 challenge;
    char user[MAX_K_NAME_SZ+1];
    int protallowed;
    int maxbufsize;
    struct sockaddr_in localaddr, remoteaddr;
    long prot_time_sec;
    char prot_time_5ms;
    /* client */
    char instance[INST_SZ];
    char realm[REALM_SZ];
    /* server */
    int (*authproc)();
    AUTH_DAT kdata;
};

/*
 * Free the space used by an opaque state pointer
 */
static void
krb_free_state(state)
void *state;
{
    memset((char *)state, 0, sizeof(struct krb_state));
    free((char *) state);
}

static int krb_is_ticket(tf_name,auth_data)
   char *tf_name; 
   struct ident_auth_data *auth_data; 
{
 struct stat stat_buf; 
 int code ; 
 char    pname[ANAME_SZ];
 char    pinst[INST_SZ];
 CREDENTIALS cred;
 time_t now;

 if (lstat(tf_name, &stat_buf) < 0) {
       if ( debug_flag && syslog_flag ) { 
	 syslog(LOG_DEBUG,"lstat error %d on %s",errno,tf_name); 
     }
 
     switch (errno) {
        case ENOENT:
            return NO_TKT_FIL;
        default:
            return TKT_FIL_ACC;
        }
 }

 /* Check ownership */ 
 if (stat_buf.st_uid != auth_data->uid ) { 
     if ( debug_flag && syslog_flag ) { 
	 syslog(LOG_DEBUG,"UID mismatch %d : %d on %s",stat_buf.st_uid,auth_data->uid,tf_name); 
     }
   return TKT_FIL_ACC;
 }     
 /* Check file perms for 600 */ 
 if ( (stat_buf.st_mode & S_IRUSR) != S_IRUSR ) {   
   if ( debug_flag && syslog_flag ) { 
	 syslog(LOG_DEBUG,"User can't read %s",tf_name); 
     }

   return TKT_FIL_ACC ; 
 }

 if ( 
    ((stat_buf.st_mode & S_IRGRP) ==  S_IRGRP) ||
    ((stat_buf.st_mode  & S_IWGRP) == S_IWGRP) ||
    ((stat_buf.st_mode  & S_IROTH) ==  S_IROTH)||
    ((stat_buf.st_mode  & S_IWOTH) ==  S_IWOTH) ) { 
   
   if (syslog_flag ) { 
     syslog(LOG_ALERT,"Incorrect file permissons on %s : %o",tf_name,stat_buf.st_mode); 
   }
 return TKT_FIL_ACC ; 
 }
 /* Okay now we should have a valid tkt file, let's look */ 
 /* Following src is C&P from klist.c in krb4. */ 

 /* Open ticket file */
 code =  tf_init(tf_name, R_TKT_FIL); 
 if ( code != KSUCCESS ) { 
   tf_close() ; 
  if ( debug_flag && syslog_flag ) { 
	 syslog(LOG_DEBUG,"tf_init failed %s",tf_name); 
     }

   return code ;
 }
 /* Get principal name and instance */
 if ((code = tf_get_pname(pname)) ||
     (code  = tf_get_pinst(pinst))) {
   tf_close () ; 
   return code ; 
 }

 /* Ensure ticket is valid */
 now = time(NULL);
 while (tf_get_cred(&cred) == KSUCCESS) {
   if (strcmp(cred.service,"krbtgt"))
     continue;
#ifdef SHORT_LIFETIME
   cred.issue_date += ((unsigned char) cred.lifetime) * 5 * 60;
#else
   cred.issue_date = krb_life_to_time(cred.issue_date, cred.lifetime);
#endif
   if (now - cred.issue_date > 1 ) {
     tf_close () ;
     return KFAILURE ;
   }
 }
 tf_close() ; 
 if ( syslog_flag ) { 
   syslog(LOG_INFO,"Ticket file %s : %s.%s for %d",tf_name,pname,pinst,auth_data->uid ); 
 }
          
 return KSUCCESS ; 
}


static int krb_find_ticket(tktstring,auth_data)
     char *tktstring; 
     struct ident_auth_data *auth_data; 
{
  int rcode ; 
  char temp[MAXPATHLEN]; 
  /* this is OS depedent */ 
  DIR *temp_dir ; 
  struct dirent *dp ; 

  /* Need to set tkt file here */ 
  /* This is cheating we need to dive kernel and get TKTFILE from ENV */ 	 /* try krb4 standard */ 
  sprintf(tktstring, "/tmp/tkt%d", auth_data->uid);

  rcode = krb_is_ticket(tktstring,auth_data) ;  
  if (rcode  == KSUCCESS ) {
    return IDENT_AUTH_OKAY ; 
  }

  /* Now we will try grupping through /tmp for a "ticket file" */ 

  temp_dir = opendir("/tmp") ; 

  for (dp = readdir(temp_dir); dp != NULL; dp = readdir(temp_dir)){
    if ( strncmp(dp->d_name, "tkt",strlen("tkt")) == 0 ) {
      if (strlen(dp->d_name) >= MAXPATHLEN-6)
	continue;
      sprintf(temp,"/tmp/%s",dp->d_name); 
      rcode = krb_is_ticket(temp,auth_data); 
      if (rcode  == KSUCCESS ) {
	strncpy(tktstring,temp,MAXPATHLEN-1); 
	closedir(temp_dir);
	return IDENT_AUTH_OKAY ; 
      }
      if (syslog_flag && debug_flag ){
	syslog(LOG_DEBUG,"Rejected %s : %d",dp->d_name,rcode); 
      }
      
    }
  }
  closedir(temp_dir);


  return IDENT_USER_CANT_AUTH ; 
}





/*
 * Start the responder side of an authentication exchange.
 */
static int krb_responder_start(auth_data)
  struct ident_auth_data *auth_data ; 
{
    struct hostent *host_name;
    char *host;
    char user[MAX_K_NAME_SZ+1];
    char instance[INST_SZ];
    char realm[REALM_SZ];
    char uinst[INST_SZ];
    char urealm[INST_SZ];
    char tktstring[MAXPATHLEN]; 
    KTEXT_ST authent;
    CREDENTIALS cr;
    struct krb_state *kstate;
    char *p;
    int i ; 

    kstate = ( struct krb_state *) malloc(sizeof(struct krb_state)) ; 
    if ( kstate == NULL )
      return IDENT_SYSTEM_ERROR ; 
    if (  krb_find_ticket(&tktstring,auth_data) != IDENT_AUTH_OKAY ) {
      return IDENT_USER_CANT_AUTH ; 
    } 
    krb_set_tkt_string(tktstring);

    /* Canonicalize hostname */
    host_name = gethostbyaddr((char *) (auth_data->remote_addr), sizeof(struct in_addr), AF_INET);
    if (!host_name) {
	return IDENT_SYSTEM_ERROR;
    }
    host = strdup(host_name->h_name);
    if (host == NULL)
        return IDENT_SYSTEM_ERROR;
    
    strncpy(instance, host_name->h_name, sizeof(instance)-1);
    instance[sizeof(instance)-1] = '\0';
    if (p = strchr(instance, '.')) *p = '\0';
    
    p = instance ; 

    while ( *p != '\0' ) { 
      if ( isupper((int) *p) ) { 
	*p = tolower((int) *p) ; 
      }
      p++ ; 
    }
  
    strncpy(realm, krb_realmofhost(host),(REALM_SZ - 1));
    free(host);

    if (syslog_flag && debug_flag ){
	syslog(LOG_DEBUG,"Fetching ticket :%s.%s@%s:",
	       IDENT_AUTH_KRB_SERVICE,instance,realm); 
    }  

    /* Fetch imap.hostname service key */
    (void) krb_mk_req(&authent,IDENT_AUTH_KRB_SERVICE , instance, realm, 0);
    memset(&authent, 0, sizeof(authent));

    if (krb_get_cred(IDENT_AUTH_KRB_SERVICE , instance, realm, &cr)) {
	return IDENT_USER_CANT_AUTH;
    }
    
    /* Do we need to know who we are ? */ 
    /*     strcpy(kstate->user, user); */ 
    
    memset((char *)kstate, 0, sizeof(*kstate));
    strcpy(kstate->service, IDENT_AUTH_KRB_SERVICE);
    kstate->authstepno = 0;
    memcpy(kstate->session, cr.session, sizeof(des_cblock));
    des_key_sched(kstate->session, kstate->schedule);


    strcpy(kstate->instance, instance);
    strcpy(kstate->realm, realm);

    memset(&cr, 0, sizeof(cr));
    auth_data->auth_method_data = (void *)kstate;
    return IDENT_AUTH_OKAY ; 
}

/*
 * Perform responder-side authentication protocol exchange
 * Returns IDENT_AUTH_OKAY if authentication can be complete after
 * sending our responder reply.
 */
static int krb_responder_auth(auth_data)
 
  struct ident_auth_data *auth_data ;
{
    static KTEXT_ST authent;
    struct krb_state *kstate = (struct krb_state *) auth_data->auth_method_data;
    char tmp[IDENT_BUFFER_SIZE];
    char resp[IDENT_BUFFER_SIZE]; 
    char output[IDENT_BUFFER_SIZE]; 
    char *resp_ptr ; 
    char line[IDENT_BUFFER_SIZE];
    char *line_ptr ; 
    IDENT_INT16 i,j,k,resp_len, len , munch; 
    KRB_INT32 challenge ; 
    struct ident_auth_data auth_reply ; 
    int code,rcode;
    int maxbufsize;
    int ready ;     
    int timeout = IDENT_READ_TIMEOUT ; 
    struct passwd *paswd_entry ;

    while ( kstate->authstepno > -1 ) { 
      switch (kstate->authstepno++) {
      case 0:
	/* Get challenge */ 
	sscanf(auth_data->authenticate_data," KERBEROS_V4 , %[A-Za-z0-9+/=]",tmp ) ; 
	rcode = ident_decodebase64(tmp); 
	
	/* Requestor gave us challenge, respond with ticket+authenticator */
	if (rcode < 4) {
	  kstate->authstepno = -1;
	  return IDENT_INVALID_REQ_INFO;
	}
	
	kstate->challenge = ntohl(*(int *)tmp);

	
	code = krb_mk_req(&authent, kstate->service, kstate->instance,
			  kstate->realm, kstate->challenge);
	if (code) {
	  kstate->authstepno = -1;
	  return IDENT_AUTH_FAIL;
	}
	/* Now we need to formulate the response */ 
	/* port, port : AUTHENTICATE : KERBEROS_V4, BASE64(htonl(authent.len)athent.dat htonl(identent.len)identent.dat) */ 
	/* the ident dat looks like 
	** [16bit flag][resp port][req port][authid len][authid][pad] 
	**  16 hton     16 hton    16 hton   16 hton     ???     pad to multiple of 8 
	** And then the whole thing is encrypted with the session key
	*/ 
	
	munch = sizeof(IDENT_INT16) ; 

	/* memcpy(to,from,length ) */ 
	k = htons(mutual_auth_flag) ; 
	resp_ptr = resp ; 
	memcpy(resp_ptr,&k,munch) ; 
	


	k = htons(auth_data->local_port) ; 
	resp_ptr += munch ; 
	memcpy(resp_ptr,&k,munch) ; 

	k = htons(auth_data->remote_port) ; 
	resp_ptr += munch ; 
	memcpy(resp_ptr,&k,munch) ; 
	

	paswd_entry = getpwuid(auth_data->uid);
	i = strlen(paswd_entry->pw_name); 

	k = htons(i) ;
	resp_ptr += munch ; 
	memcpy(resp_ptr,&k,munch) ;
	resp_ptr += munch ;

	strcpy(resp_ptr,paswd_entry->pw_name);

	k = sizeof(des_cblock) ; 
	j = i%k; 
	resp_ptr += i ; 

	/* Pad to a multiple of sizeof(des_cblock)  */ 
	bzero(resp_ptr,k-j) ; 
	resp_len = 4*(munch)+i+k-j ;
	
	if ( syslog_flag && debug_flag ) { 
	  syslog(LOG_DEBUG,"Data: %4.4x:%4.4x:%4.4x:%4.4x:%s",
		 htons(mutual_auth_flag) ,
		 htons(auth_data->remote_port),
		 htons(auth_data->local_port),
		 htons(i),
		 paswd_entry->pw_name);
	  syslog(LOG_DEBUG,"Resp:%d:%4.4x:%4.4x:%4.4x:%4.4x:%s",munch,
		 *((IDENT_INT16*) resp),
		 *((IDENT_INT16*) &(resp[munch])),
		 *((IDENT_INT16*) &(resp[2*munch])),
		 *((IDENT_INT16*) &(resp[3*munch])),
		 &(resp[4*munch]));
	}
	des_pcbc_encrypt((des_cblock *) resp,(des_cblock *) output, resp_len ,kstate->schedule,&(kstate->session),1 ) ; 
	
	resp_ptr = tmp ; 
        k = htons(authent.length); 
	memcpy(resp,(char*)&k,(munch)); 
	resp_ptr = &(resp[(munch)]); 
	memcpy(resp_ptr,authent.dat,authent.length); 
	resp_ptr = &(resp[(munch)+authent.length]); 
	k = htons(resp_len); 
	memcpy(resp_ptr,(char *)&k,(munch)); 
	resp_ptr = &(resp[(munch)+authent.length+(munch)]);
	memcpy(resp_ptr,output,resp_len); 
	len = munch + authent.length + munch + resp_len ;
 
	ident_writebase64(resp,tmp,len); 
	sprintf(line," AUTHENTICATE : %s , %s ",auth_data->auth_struct->auth_method,tmp ); 
	do_out(line,auth_data); 

	if (mutual_auth_flag & MUTUAL_AUTH_BIT ) { 
	  return IDENT_AUTH_OKAY;
	}
	break;
	
      case 1:
	/*
	 * Requestor gave us mutual auth reply.
	 */
	line[0] = '\0' ; 
	
	line_ptr = sfgets(line,IDENT_BUFFER_SIZE,auth_data->fp,timeout,auth_data);
	if ( line_ptr == NULL ) {
	  do_error(IDENT_NO_MUTUAL_AUTH,auth_data); 
	}
	rcode = parse_auth_line(line,&auth_reply);
	if( rcode != IDENT_AUTH_OKAY ) { 
	  do_error(rcode,&auth_reply) ; 
	}
	/* Check if the data is the same */ 
	if (
	    ( auth_reply.local_port != auth_data->local_port ) || 
	    ( auth_reply.remote_port != auth_data->remote_port ) || 
	    ( strncmp(auth_reply.auth_struct->auth_method ,
		      auth_data->auth_struct->auth_method , 
		      strlen(auth_data->auth_struct->auth_method)) != 0 ) ) { 

	  if (debug_flag && syslog_flag)
	    syslog(LOG_DEBUG, "Auth reply does not match auth_data %d,%d : %d,%d", auth_reply.local_port,auth_reply.remote_port,auth_data->local_port,auth_data->remote_port );
	  do_error(IDENT_MUTUAL_AUTH_FAIL,auth_data); 
	}
	/* Get challenge */ 
	sscanf(auth_reply.authenticate_data," KERBEROS_V4 , %[A-Za-z0-9+/=]",tmp ) ; 
	rcode = ident_decodebase64(tmp); 

	if (rcode < 4) {
	  kstate->authstepno = -1;
	  return IDENT_INVALID_RESP_INFO;
	}
	
	des_ecb_encrypt(tmp, tmp, kstate->schedule, 0);
	memcpy(&challenge,tmp,sizeof(challenge)) ; 
	challenge = ntohl(challenge); 
	if (challenge != kstate->challenge + 1) {
	    /* Server failed to mutually authenticte */
	    kstate->authstepno = -1;
	if (debug_flag && syslog_flag)
	  syslog(LOG_DEBUG, "Challenge failed:%d:%d",kstate->challenge,challenge);
	    return IDENT_MUTUAL_AUTH_FAIL;
	}	    
	if (debug_flag && syslog_flag)
	  syslog(LOG_DEBUG, "Challenge succeeded:%d:%d",kstate->challenge,challenge);
	
	return IDENT_AUTH_OKAY;

      default:
	kstate->authstepno = -1;
	return IDENT_SYSTEM_ERROR;
      }
    }
}


/*
** Does trivial (i.e. passwd file based auth)
**
*/
static int trivial_responder_start( auth_data)
struct ident_auth_data *auth_data; 
{ 

  struct passwd *paswd_entry ; 
  if ( allow_trivial ) { 
    paswd_entry = getpwuid(auth_data->uid);
    if (!paswd_entry || auth_data->uid != paswd_entry->pw_uid) { 
      do_error(IDENT_NO_USER,auth_data);
    }
    auth_data->auth_method_data = (void *) paswd_entry ; 
    return IDENT_AUTH_OKAY ; 
  } else { 
    do_error(IDENT_INVALID_REQ_INFO,auth_data);  
  }

} 

static int trivial_responder_auth( auth_data)
struct ident_auth_data *auth_data; 
{ 
  struct passwd *pwd_entry ;
  char tmp[IDENT_BUFFER_SIZE] ; 
  if ( allow_trivial) { 
    /* Do we want to do noident check here ? */ 
    pwd_entry = ( struct passwd *) auth_data->auth_method_data;  
    if ( ! uidonly_flag ) { 
    sprintf(tmp," USERID : UNIX :%s ",
	      pwd_entry->pw_name);
    } else { 
      sprintf(tmp," USERID : UNIX :%d ",
	      auth_data->uid) ; 
    }
    do_out(tmp,auth_data); 
    return IDENT_AUTH_OKAY ; 
  } else { 
    do_error(IDENT_INVALID_REQ_INFO,auth_data); 
  }
} 






static int  null_responder_start() { 
  return IDENT_AUTH_NOT_SUPPORTED ; 

}

static int     null_responder_auth() { 

  return IDENT_AUTH_NOT_SUPPORTED ; 
}

static void     null_free_state() {
  return ; 
}
