/** @file
 * The S/Ident protocol parser.
 *
 * @author Peter Eriksson <pen@lysator.liu.se>
 * @author with modifications for S/Ident by Booker C. Bense
 *  <bbense@stanford.edu>
 *
 * $Id: parse.c,v 1.8 2002/10/29 00:52:15 eagle Exp $
 */

#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

/* BSDI needs <netinet/in.h> before <arpa/inet.h>. */
#include <netinet/in.h>
#include <arpa/inet.h>

#include "error.h"
#include "identd.h"
#include "sident.h"
#include "pidentd.h" 

/*
** This function will eat whitespace characters until
** either a non-whitespace character is read, or EOF
** occurs. This function is only used if the "-m" option
** is enabled.
*/
static int eat_whitespace()
{
  int c;


  while ((c = getchar()) != EOF &&
	 !(c == '\r' || c == '\n'))
    ;

  if (c != EOF)
    while ((c = getchar()) != EOF &&
	   (c == ' ' || c == '\t' || c == '\n' || c == '\r'))
      ;

  if (c != EOF)
    ungetc(c, stdin);

  return (c != EOF);
}



/*
** A small routine to check for the existance of the ".nosident"
** file in a users home directory.
*/
static int check_nosident(uid_t uid)
{
  char *tmp_path,*homedir;
  struct stat sbuf;
  int rcode;
  struct passwd *pw ; 

  if ( users_can_nosident ) { 

    pw = getpwuid(uid); 
    
    if ( pw == NULL ) 
      return 0 ; 

    homedir = pw->pw_dir ; 

    if (!homedir)
      return 0;

    tmp_path = (char *) malloc(strlen(homedir) + sizeof("/.nosident") + 1);
    if (!tmp_path)
      return 0;

    strcpy(tmp_path, homedir);
    strcat(tmp_path, "/.nosident");

    rcode = stat(tmp_path, &sbuf);
    free(tmp_path);

    return (rcode == 0);
  } else { 
    return ( 1 == 0 ) ; 
  }
}


/*
**  Borrowed from Sendmail.8.8 
**  SFGETS -- "safe" fgets -- times out and ignores random interrupts.
**
**	Parameters:
**		buf -- place to put the input line.
**		siz -- size of buf.
**		fp -- file to read from.
**		timeout -- the timeout before error occurs.
**		during -- what we are trying to read (for error messages).
**
**	Returns:
**		NULL on error (including timeout).  This will also leave
**			buf containing a null string.
**		buf otherwise.
**
**	Side Effects:
**		none.
*/

static jmp_buf	CtxReadTimeout;
static void	readtimeout();

char *
sfgets(buf, siz, fp, timeout,auth_data)
	char *buf;
	int siz;
	FILE *fp;
	time_t timeout;
        struct ident_auth_data *auth_data ;
{

	register char *p;

	if (fp == NULL)
	{
		buf[0] = '\0';
		return NULL;
	}

	/* set the timeout */
	if (timeout != 0)
	{
	  if (signal(SIGALRM, readtimeout) == SIG_ERR) 
	          do_error(IDENT_SYSTEM_ERROR,auth_data) ; 
	  if (setjmp(CtxReadTimeout) != 0)
	    {
	      syslog(LOG_NOTICE,
		     "timeout waiting for input");
	      errno = 0;

	      buf[0] = '\0';
	      return (NULL);
	    }
	  alarm(timeout); 
	}

	/* try to read */
	p = NULL;
	while (!feof(fp) && !ferror(fp))
	{
		errno = 0;
		p = fgets(buf, siz, fp);
		if (p != NULL || errno != EINTR)
			break;
		clearerr(fp);
	}

	/* clear the event if it has not sprung */

	alarm(0); 

	if (verbose_flag)
	  syslog(LOG_DEBUG, "In function sfgets :%d:%s:%s",siz,buf,p);

	return (buf);
}

static void
readtimeout(sig_no)
	int sig_no;
{
	longjmp(CtxReadTimeout, 1);
}

/*
** Parse incoming lines 
**
*/
int 
parse_auth_line(line,auth_data) 
char *line; 
struct ident_auth_data *auth_data ;
{ 
  char arg[33] ;
  int rcode ; 
  char *p ; 
  char delim[] = ":" ; 

  p = strtok(line,delim); 
  if (verbose_flag)
    syslog(LOG_DEBUG, "In function parse_auth_line:%s:port_data %s",line,p);

  if ( p != NULL ) { 
    auth_data->port_data = p ; 
  } else { 
    return IDENT_INVALID_PORT ; 
  }

  p = strtok((char *) NULL,delim); 
  auth_data->cmd_data = p ; 
  
  p = strtok((char *) NULL,delim); 
  auth_data->authenticate_data = p ; 
  
  if ( auth_data->port_data != NULL ) { 
    rcode = sscanf(auth_data->port_data, " %d , %d ", &(auth_data->local_port),
		   &(auth_data->remote_port));
  } else { 
    return IDENT_INVALID_PORT ; 
  }
  if ( rcode > 0 ) { 
    if (auth_data->cmd_data != NULL ) { 
      rcode = sscanf(auth_data->cmd_data," %32s ",arg ) ;
    } else {
      rcode = 0 ; 
    }
    if ( rcode > 0 ) { 
      if (( strncmp(arg,"AUTHENTICATE",strlen("AUTHENTICATE")) == 0)
	  && (auth_data->authenticate_data != NULL)  ) {
        if (verbose_flag)
          syslog(LOG_DEBUG, "About to parse %s", auth_data->authenticate_data);
	rcode = parse_auth(auth_data->authenticate_data,&(auth_data->auth_struct)) ; 
	return rcode ; 
      } else { 
	return IDENT_INVALID_REQ_INFO ; 
      }
    } else if ( allow_trivial   ) {
       rcode =  parse_auth(NULL,&(auth_data->auth_struct)) ; 
       return rcode ; 
    } else { 
      return IDENT_UNKNOWN_ERROR ;
    } 
  } else { 
    return IDENT_SYSTEM_ERROR; 
  }
  
}
int 
do_out(string , auth_data)
char *string ; 
struct ident_auth_data *auth_data ; 
{
  printf(" %d , %d : %s\r\n", auth_data->local_port,auth_data->remote_port,string); 
  if (verbose_flag)
    syslog(LOG_DEBUG," %d , %d : %s", auth_data->local_port,auth_data->remote_port,string );
  
   fflush(stdout); 
  
} 
/* 
** Handle ERROR codes 
**
*/
int 
do_error(e_code,auth_data) 
  int e_code ; 
  struct ident_auth_data *auth_data ;
{ 
  char tmp[2048] ; 
  int show_code ; 
  if ( e_code < IDENT_INTERNAL_ERR ) { 
    if ( unknown_flag  ) {
      if ( e_code == IDENT_NO_USER || e_code == IDENT_INVALID_PORT ) 
	show_code = IDENT_UNKNOWN_ERROR ; 
    } else { 
      show_code = e_code ; 
    }
    sprintf(tmp," ERROR : %s ", ident_err_txt[show_code]);
    do_out(tmp,auth_data); 

    if ( syslog_flag)
      syslog(LOG_INFO, "Error : %s", ident_err_txt[e_code]);
      
  } else { 
    /* Need to syslog errors here */ 
    switch(e_code) { 
    case IDENT_NO_MUTUAL_AUTH: 
      if ( syslog_flag) 
	syslog(LOG_INFO, " No mutual auth data");
      break;
    case IDENT_MUTUAL_AUTH_FAIL: 
      if ( syslog_flag)
	syslog(LOG_INFO, " Mutual auth failed" );
      break ;
    case IDENT_SYSTEM_ERROR: 
      if ( syslog_flag)
	syslog(LOG_INFO, " System error" );
      break ;

    default:
      if (syslog_flag)
	syslog(LOG_INFO, " Unknown error code %d",e_code );
      ; 
    } 
  } 
  exit(e_code); 

}  



int parse(fp, laddr, faddr)
  FILE *fp;
  struct in_addr *laddr, *faddr;
{
  int try, rcode;
  int timeout = IDENT_READ_TIMEOUT ; 

  struct ident_auth_data auth_data ; 
  
  struct in_addr laddr2;
  struct in_addr faddr2;
  int k_opened;
  char line[2048]; 
  char *line_ptr ; 
  void *ka_ptr ; 
  struct  kernel *kinfo_ptr ; 

  k_opened = 0;

  if (verbose_flag)
    syslog(LOG_DEBUG, "In function parse()");

  /*
  ** Get the local/foreign port pair from the luser
  */

  if (verbose_flag)
    syslog(LOG_DEBUG, "  Before sfgets()");

  faddr2 = *faddr;
  laddr2 = *laddr;
 
  auth_data.local_addr = &laddr2;
  auth_data.remote_addr = &faddr2; 
  auth_data.fp = fp ; 

  auth_data.local_port = auth_data.remote_port = 0;
   
  /* Read query from client */
  line_ptr = line ;
 
  line_ptr = sfgets(line,2048,fp,timeout,&auth_data); 
  if (verbose_flag)
    syslog(LOG_DEBUG, "After sfgets line_ptr:%s",line_ptr);

  if ( line_ptr == NULL  ) { 
    do_error(IDENT_SYSTEM_ERROR,&auth_data); 
  } 
  rcode = parse_auth_line(line_ptr,&auth_data); 
  if ( rcode != IDENT_AUTH_OKAY ) { 
    do_error(rcode,&auth_data); 
  }


  if (!( VALID_PORT(auth_data.local_port) &&  VALID_PORT(auth_data.remote_port))) { 
    do_error(IDENT_INVALID_PORT,&auth_data); 
  }
    
  if ( client_only_flag ) { 
    if( SERVER_PORT(auth_data.local_port) ) { 
      do_error(IDENT_NO_USER,&auth_data); 
    }
  }

  if (syslog_flag) {
    syslog(LOG_NOTICE, "from: %s ( %s ) for: %d, %d",
	   inet_ntoa(*faddr), gethost(faddr), auth_data.local_port, auth_data.remote_port);
  }

  if (verbose_flag)
    syslog(LOG_DEBUG, "  After fscanf(), before k_open()");

  if (! k_opened) {
    /*
    ** Open the kernel memory device and read the nlist table
    ** 
    ** Converted to use the ka_open routines from pidentd.3.0.x
    */
    if (ka_open(&ka_ptr) != 0) {
      if (syslog_flag) 
	syslog(LOG_ERR, "ka_open call failed");
      do_error(IDENT_UNKNOWN_ERROR,auth_data);
      
    }
    k_opened = 1;
  }

  if (verbose_flag)
    syslog(LOG_DEBUG, "  After ka_open(), before ka_lookup()");

  /*
  ** Get the specific TCP connection and return the uid - user number.
  */

#define MAX_RETRY 20
  /*
  ** Try to fetch the information MAX_RETRY times in case the
  ** kernel changed beneath us and we missed or took a fault.
  **
  ** Why would we ever fail? Is not there a reliable way for the
  ** kernel to identify its sockets? Cannot we use that interface?
  **
  ** Used to be 5 times, but often this is not enough on Alpha OSF.
  */
  /* #define SLEEP_BETWEEN_RETRIES 1 */
  /*
  ** If we failed in k_getuid, that is presumably because the OS was
  ** busy creating or destroying processes. We may want to sleep for
  ** a random time between retries, hoping for peace and quiet.
  */

  /* Fill up kernel info struct */ 

  /*  struct kernel */
  /*  { */
  /*       Request parameters  */
  /*      struct sockaddr_in local; */
  /*      struct sockaddr_in remote; */
    
  /*      Result parameters */
  /*      avail_t av; */
    
  /*      int status; */
  /*      uid_t euid; */
  /*      uid_t ruid; */
    
  /*      pid_t pid; */
  /*      char *cmd; */
  /*      char *argv; */
  /*  }; */

  kinfo_ptr = kernel_alloc() ; 

  if ( kinfo_ptr == NULL ) { 
    if (syslog_flag) 
	syslog(LOG_ERR, "kernel_alloc call failed");
    do_error(IDENT_UNKNOWN_ERROR,auth_data);
  }

  kinfo_ptr->local.sin_port = htons(auth_data.local_port) ; 
  kinfo_ptr->remote.sin_port  = htons(auth_data.remote_port) ;
  memcpy(&(kinfo_ptr->local.sin_addr),auth_data.local_addr,sizeof(struct in_addr)); 
  memcpy(&(kinfo_ptr->remote.sin_addr),auth_data.remote_addr,sizeof(struct in_addr)); 
  kinfo_ptr->local.sin_family = AF_INET ; 
  kinfo_ptr->remote.sin_family = AF_INET ; 
  /* ka_lookup returns 0 on not-found,1 on found, -1 on failure. */
  kinfo_ptr->status = -1 ; 

  for (try = 0;(try < MAX_RETRY && kinfo_ptr->status <  0); try++) {
    kinfo_ptr->status =  ka_lookup(ka_ptr,kinfo_ptr) ; 
      
#ifdef SLEEP_BETWEEN_RETRIES
    if (  kinfo_ptr->status <  0 ) {  
      /* Seed the generator: auth_data.local_port should be unique (among other concurrent identd's) */
      if (try < 1) srandom(auth_data.local_port);
      /* This gives a max sleep of 0xffff = 65535 microsecs, about 32millisec average */
      usleep(random()&0x00ffff);
    }
#endif
  }
  if (try >= MAX_RETRY) {
    if (syslog_flag)	
      syslog(LOG_INFO, "MAX_RETRY exceeded,Returned: %d , %d : NO-USER", auth_data.local_port, auth_data.remote_port);
    do_error(IDENT_NO_USER,&auth_data); 
  }

  if (try > 1 && verbose_flag)
    syslog(LOG_DEBUG, "ka_lookup retries: %d", try);

  if (verbose_flag)
    syslog(LOG_DEBUG, "  After ka_lookup(), before getpwuid() kinfo_ptr status = %d",kinfo_ptr->status);

  if ( kinfo_ptr->status == 0 ) { 
    do_error(IDENT_NO_USER,&auth_data); 
  }

  if (kinfo_ptr->ruid == NO_UID)
    auth_data.uid = kinfo_ptr->euid;
  else
    auth_data.uid = kinfo_ptr->ruid;
 

  if ( auth_data.uid > -1 ) { 
      /* Setuid here ??? */ 

      if ( setuid(auth_data.uid) != 0 ) { 
	  syslog(LOG_ALERT,"Setuid failed for uid:%d",auth_data.uid); 
	  do_error(IDENT_UNKNOWN_ERROR,&auth_data); 
      }
      /* Check for .nosident */ 
      if( check_nosident(auth_data.uid) ) { 
	  syslog(LOG_DEBUG,"No ident requested by uid:%d",auth_data.uid) ; 
	  do_error(IDENT_HIDDEN_USER,&auth_data); 
      } 
  } else { 
    if (syslog_flag)
      syslog(LOG_INFO, "INVALID uid %d",auth_data.uid);
    do_error(IDENT_UNKNOWN_ERROR,&auth_data); 
  }

  if (syslog_flag)
    syslog(LOG_INFO, "AUTHENTICATE with %s",auth_data.auth_struct->auth_method);

  rcode = (*auth_data.auth_struct->start)(&auth_data);
  if ( rcode != IDENT_AUTH_OKAY ) { 
    if (verbose_flag)
      syslog(LOG_DEBUG, "  Fail in auth_start");
    do_error(rcode,&auth_data);
  }

  rcode = (*auth_data.auth_struct->auth)(&auth_data);
  if ( rcode != IDENT_AUTH_OKAY ) { 
    if (verbose_flag)
      syslog(LOG_DEBUG, " Fail in auth_auth");
    do_error(rcode,&auth_data); 
  }     
  return 0;
}
